library docs_analyzer;

import 'dart:async';
import 'dart:io';

import 'package:analyzer/file_system/file_system.dart' as fileSystem;
import 'package:analyzer/file_system/physical_file_system.dart';
import 'package:analyzer/source/package_map_resolver.dart';
import 'package:analyzer/source/pub_package_map_provider.dart';
import 'package:analyzer/source/sdk_ext.dart';
import 'package:analyzer/src/generated/element.dart';
import 'package:analyzer/src/generated/engine.dart';
import 'package:analyzer/src/generated/java_io.dart';
import 'package:analyzer/src/generated/sdk.dart';
import 'package:analyzer/src/generated/sdk_io.dart';
import 'package:analyzer/src/generated/source.dart';
import 'package:analyzer/src/generated/source_io.dart';
import 'package:dartdoc/dartdoc.dart';
import 'package:dartdoc/src/io_utils.dart';
import 'package:dartdoc/src/model.dart';
import 'package:dartdoc/src/package_meta.dart';

/// Analyze documentation for the library in the [inputDir]. [sdkDir] is the location of the Dart
/// SDK.
Future<DocsAnalysis> analyzeDocumentation(Directory inputDir, Directory sdkDir) async {
  PackageMeta packageMeta = new PackageMeta.fromDir(inputDir);
  DartDoc doc = new DartDoc(inputDir, sdkDir, packageMeta);
  Package results = await doc.generateDocs();

  return new DocsAnalysis(results);
}

/// Collect Dart documentation for all public Dart libraries in the given directory.
class DartDoc {
  final Directory rootDir;
  final Directory sdkDir;
  final PackageMeta packageMeta;

  DartDoc(this.rootDir, this.sdkDir, this.packageMeta);

  /// Generate DartDoc documentation.
  ///
  /// [DartDocResults] is returned if dartdoc succeeds. If analysis failed due to an error in the
  /// analyzed project, prints a message and exits.
  Future<Package> generateDocs() async {
    List<String> files = findFilesToDocumentInPackage(rootDir.path).toList();
    List<LibraryElement> libraries = _parseLibraries(files);
    return new Package(libraries, packageMeta);
  }

  List<LibraryElement> _parseLibraries(List<String> files) {
    Set<LibraryElement> libraries = new Set();
    DartSdk sdk = new DirectoryBasedDartSdk(new JavaFile(sdkDir.path));
    List<UriResolver> resolvers = [new DartUriResolver(sdk)];

    fileSystem.Resource cwd = PhysicalResourceProvider.INSTANCE.getResource(rootDir.path);

    var pubPackageMapProvider = new PubPackageMapProvider(PhysicalResourceProvider.INSTANCE, sdk);

    var packageMapInfo = pubPackageMapProvider.computePackageMap(cwd);

    Map<String, List<fileSystem.Folder>> packageMap = packageMapInfo.packageMap;
    if (packageMap != null) {
      resolvers.add(new SdkExtUriResolver(packageMap));
      resolvers.add(new PackageMapUriResolver(PhysicalResourceProvider.INSTANCE, packageMap));
    }
    resolvers.add(new FileUriResolver());

    SourceFactory sourceFactory = new SourceFactory(resolvers);

    var options = new AnalysisOptionsImpl()..cacheSize = 512;

    AnalysisContext context = AnalysisEngine.instance.createAnalysisContext()
      ..analysisOptions = options
      ..sourceFactory = sourceFactory;

    List<Source> sources = [];

    files.forEach((String filePath) {
      String name = filePath;
      if (name.startsWith(Directory.current.path)) {
        name = name.substring(Directory.current.path.length);
        if (name.startsWith(Platform.pathSeparator)) name = name.substring(1);
      }

      JavaFile javaFile = new JavaFile(filePath);
      Source source = new FileBasedSource(new JavaFile(filePath));
      Uri uri = context.sourceFactory.restoreUri(source);
      if (uri != null) {
        source = new FileBasedSource(javaFile, uri);
      }
      sources.add(source);
      if (context.computeKindOf(source) == SourceKind.LIBRARY) {
        LibraryElement library = context.computeLibraryElement(source);
        libraries.add(library);
      }
    });

    // Ensure that the analysis engine performs all remaining work.
    AnalysisResult result = context.performAnalysisTask();
    while (result.hasMoreWork) {
      result = context.performAnalysisTask();
    }

    for (Source source in sources) {
      context.computeErrors(source);
      var errors = context.getErrors(source).errors;

      if (!errors.isEmpty) {
        print('Errors were encountered during analysis. Try running the dart analyzer.');
        exit(1);
      }
    }

    return libraries.toList();
  }
}

/// Analyzes a [Package] for missing documentation.
class DocsAnalysis {
  final List<ModelElement> allElements = [];
  final List<ModelElement> undocumentedElements = [];

  final List<Class> classes = [];
  final List<Class> enums = [];
  final List<TopLevelVariable> variables = [];
  final List<ModelFunction> functions = [];
  final List<Typedef> typedefs = [];
  final List<Class> exceptions = [];

  final List<Class> undocumentedClasses = [];
  final List<Class> undocumentedEnums = [];
  final List<TopLevelVariable> undocumentedVariables = [];
  final List<ModelFunction> undocumentedFunctions = [];
  final List<Typedef> undocumentedTypedefs = [];
  final List<Class> undocumentedExceptions = [];

  bool get isMissingDocumentation => undocumentedElements.length != 0;

  DocsAnalysis(Package package) {
    package.libraries.forEach((lib) {
      if (lib.documentation.contains('@nodoc')) {
        return;
      }

      classes.addAll(lib.allClasses);
      enums.addAll(lib.enums);
      variables.addAll(lib.properties);
      variables.addAll(lib.constants);
      functions.addAll(lib.functions);
      typedefs.addAll(lib.typedefs);
      exceptions.addAll(lib.exceptions);
    });

    allElements.addAll([classes, enums, variables, functions, typedefs].expand((l) => l));
    undocumentedElements.addAll(allElements.where(_elementLacksDocumentation));

    undocumentedClasses.addAll(classes.where(_elementLacksDocumentation));
    undocumentedEnums.addAll(enums.where(_elementLacksDocumentation));
    undocumentedVariables.addAll(variables.where(_elementLacksDocumentation));
    undocumentedFunctions.addAll(functions.where(_elementLacksDocumentation).toList());
    undocumentedTypedefs.addAll(typedefs.where(_elementLacksDocumentation).toList());
    undocumentedExceptions.addAll(exceptions.where(_elementLacksDocumentation));
  }

  DocsAnalysis.fromResult(DartDocResults results) : this(results.package);

  bool _elementLacksDocumentation(ModelElement elem) {
    return !elem.hasDocumentation;
  }
}
